(* fonction utilitaire *)

(** Étant donné un fichier texte de nom f, renvoie la liste de ses lignes.
    Chaque ligne est une chaîne de caractères sans \n final.
*)
let read_file f =
  let cin = open_in f in
  Fun.protect
    ~finally:(fun () -> close_in cin)
    (fun () -> String.split_on_char '\n' (In_channel.input_all cin))

module IntSet = Tree.Make (Int)
(** Q1.2 *)

let iset = IntSet.(add 42 (add 18 (add 191 empty)))

(** Q1.4*)
let () = Printf.printf "%s\n" (IntSet.to_string iset)

(** Q1.5*)
let subset_without_duplicates l =
  let rec loop l seen gacc cacc =
    (* seen: on a vu un doublon,
        gacc: accmulateur global
        lacc: accumulateur courant *)
    match l with
    | [] -> gacc
    | "--" :: ll -> loop ll false (IntSet.union gacc cacc) IntSet.empty
    | s :: ll ->
        if seen then loop ll true gacc IntSet.empty
        else
          let i = int_of_string s in
          if IntSet.mem i cacc then loop ll true gacc IntSet.empty
          else loop ll false gacc (IntSet.add i cacc)
  in
  loop l false IntSet.empty IntSet.empty

let test_1_5 file =
  let lines1 = read_file file in
  Printf.printf "Q1.5 %s\n"
    (IntSet.to_string (subset_without_duplicates lines1))

(** pas demandé
let () =
for _ = 0 to 100 do
  for _ = 0 to Random.int 200 do
    Printf.printf "%d\n" (Random.int 10_000)
  done;
  Printf.printf "--\n"
done
*)
let () =
  test_1_5 "intset1.txt";
  test_1_5 "intset2.txt"

(* Q1.6*)
module Chr = struct
  type t = char

  let compare = Char.compare
  let to_string c = "'" ^ Char.escaped c ^ "'"
end

module ChrSet = Tree.Make (Chr)

(* Q1.7*)
let charset s = String.fold_left (fun acc c -> ChrSet.add c acc) ChrSet.empty s
let row1 = charset "qwertyuiop"
let row2 = charset "asdfghjkl"
let row3 = charset "zxcvbnm"

(* Q1.8 *)
let only_one_row l =
  let rec loop l acc =
    match l with
    | [] -> acc
    | s :: ll ->
        let s = String.lowercase_ascii s in
        let cs = charset s in
        let acc =
          if ChrSet.subset cs row1 then (1, s) :: acc
          else if ChrSet.subset cs row2 then (2, s) :: acc
          else if ChrSet.subset cs row3 then (3, s) :: acc
          else acc
        in
        loop ll acc
  in
  loop l []

let test_1_8 f =
  let lines = read_file f in
  let words = only_one_row lines in
  List.iter (fun (i, w) -> Printf.printf "%s (row %d)\n" w i) words

let () = test_1_8 "english.txt"

(* Q2.3 tests pour remove_min et remove_max *)
let test_2_3 () =
  let set = subset_without_duplicates (read_file "intset2.txt") in
  (* crée un grand arbre *)
  let set = IntSet.(add min_int (add max_int set)) in
  (* ajoute le plus petit et plus grand entier *)
  let seti, inf = IntSet.remove_min set in
  let setf, sup = IntSet.remove_max seti in
  assert (inf = min_int);
  assert (sup = max_int);
  assert (not (IntSet.mem max_int setf));
  assert (not (IntSet.mem min_int seti))

let () = test_2_3 ()

(* Q2.4 *)
let smallest_such_than p s =
  let it = IntSet.mk_iterator s in
  let rec loop it =
    match IntSet.next it with
    | None -> None
    | Some (v, itt) -> if p v then Some v else loop itt
  in
  loop it

let test_2_4 () =
  let set = subset_without_duplicates (read_file "intset2.txt") in
  Printf.printf "Plus petit multiple non nul de 100 dans l'ensemble: ";
  match smallest_such_than (fun x -> x <> 0 && x mod 100 = 0) set with
  | None -> Printf.printf "aucun !\n"
  | Some i -> Printf.printf "%d\n" i

let () = test_2_4 ()